"""
noseperf.testcases
~~~~~~~~~~~~~~~~~~

:copyright: (c) 2012 DISQUS
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import

from unittest2 import TestCase

__all__ = ('PerformanceTest', 'DjangoPerformanceTest')


class PerformanceTest(TestCase):
    """
    A base class for performance tests.

    Nose can discover these classes and execute the tests. The performance
    plugin will then profile the tests and measure activity.
    """
    tags = ["performance"]

try:
    __import__('django')
except ImportError:
    class DjangoPerformanceTest(PerformanceTest):
        pass  # NOOP
else:
    from django.test import TestCase as DjangoTestCase

    class DjangoPerformanceTest(DjangoTestCase, PerformanceTest):
        pass



"""
noseperf.plugin
~~~~~~~~~~~~~~~

:copyright: (c) 2012 DISQUS
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import

from datetime import datetime
import json
import os
import sys
import time

from nose.plugins.base import Plugin

from noseperf.testcases import PerformanceTest
from noseperf.util import PatchContext
from noseperf.wrappers.base import FunctionWrapper


class PerformancePlugin(Plugin):
    """
    Nose performance plugin.

    This plugin is used with the PerformanceTest class. This plugin will
    profile each test and dump all of the raw data into a json file.

    For this test to be super helpful, pass in a fixture set generated by the
    ``loaddata`` command. Then run the tests as follows::

        nosetests /tests/performance/ --with-performance

    """
    score = -sys.maxint - 1  # Should be the absolute last thing ever to run since it transparently monitors the code.
    enabled = False
    version = 2

    def options(self, parser, env):
        parser.add_option("--with-performance", dest="enable_performance",
            default=False, action="store_true",
            help="Enable the performance plugin.")

        parser.add_option("--performance-json-file", dest="performance_file",
            default="performance.json", metavar="FILE",
            help="Save Results in file. Defaults to performance.json")

        parser.add_option("--performance-revision", dest="performance_revision",
            default=None, help="Revision name to identify this build")

        parser.add_option("--performance-schema", dest="performance_schema",
            default=None, help="Schema to identify this build with")

        parser.add_option("--disable-performance-selector", dest="disable_performance_selector",
            action='store_true', help="If disabled, will allow any test to run", default=False)

    def configure(self, options, conf):
        if not options.enable_performance:
            return

        self.enabled = True
        self.json_file = options.performance_file
        self.revision = options.performance_revision
        self.schema = options.performance_schema
        self.use_selector = not options.disable_performance_selector

    def wantClass(self, cls):
        """
        Only profile performance tests.
        """
        if not self.use_selector:
            return None
        return issubclass(cls, PerformanceTest)

    def begin(self):
        self.start_time = datetime.now()
        self.tests = []
        self.context_stack = []

    def patch_interfaces(self):
        self._calls = []

        try:
            from django.conf import settings
            if not settings.configured:
                raise ImportError
        except ImportError:
            pass
        else:
            self.patch_django_interfaces()

        try:
            __import__('redis')
        except ImportError:
            pass
        else:
            self.patch_redis_interfaces()

    def patch_redis_interfaces(self):
        from noseperf.wrappers.redis import RedisWrapper, RedisPipelineWrapper

        self.add_context(PatchContext('redis.client.StrictRedis.execute_command', RedisWrapper(self._calls)))
        self.add_context(PatchContext('redis.client.BasePipeline.execute', RedisPipelineWrapper(self._calls)))

    def patch_django_interfaces(self):
        import django
        from django.conf import settings
        from noseperf.wrappers.django import DjangoTemplateWrapper, patch_cursor

        self.add_context(PatchContext('django.db.backends.BaseDatabaseWrapper.cursor', patch_cursor(self._calls)))

        self.add_context(PatchContext('django.template.Template._render', DjangoTemplateWrapper(self._calls)))

        cache_backends = set()
        if django.VERSION < (1, 3):
            backend = settings.CACHE_BACKEND.split(':', 1)[0]
            if '.' not in backend:
                backend = 'django.core.cache.backends.%s' % backend
            path = '%s.CacheClass' % backend
            cache_backends.add(path)

        else:
            for cache_config in settings.CACHES.itervalues():
                cache_backends.add(cache_config['BACKEND'])

        for path in cache_backends:
            for cmd in ('get', 'set', 'add', 'delete', 'get_many'):
                self.add_context(PatchContext('%s.%s' % (path, cmd), FunctionWrapper('cache', self._calls)))

    def add_context(self, ctx):
        ctx.__enter__()
        self.context_stack.append(ctx)

    def clear_context(self):
        while self.context_stack:
            self.context_stack.pop().__exit__(None, None, None)

    def beforeTest(self, test):
        """
        Monkey patch the test to ensure setUp and tearDown happen outside
        of the scope of our monitoring hooks.
        """
        plugin = self

        cls = type(test.test)

        cls.__orig_setup = cls.setUp
        cls.__orig_teardown = cls.tearDown

        def setUp(self):
            self.__orig_setup()
            plugin.beginTracing(self)

        def tearDown(self):
            plugin.endTracing(self)
            self.__orig_teardown()

        setUp.__is_patched = True
        tearDown.__is_patched = True

        cls.setUp = setUp
        cls.tearDown = tearDown

        return test

    def afterTest(self, test):
        """
        Remove our monkey patches.
        """
        cls = type(test.test)

        cls.setUp = cls.__orig_setup
        cls.tearDown = cls.__orig_teardown

    def beginTracing(self, test):
        self.patch_interfaces()
        self.start = time.time()

    def endTracing(self, test):
        self.end = time.time()
        self.clear_context()

        if not hasattr(self, 'end'):
            return

        interfaces = {}

        data = {
            'id': test.id(),
            'doc': test._testMethodDoc,
            'duration': self.end - self.start,
            'interfaces': interfaces,
            'calls': self._calls,
        }

        self.tests.append(data)

    def report(self, stream):
        json_file = os.path.realpath(self.json_file)
        base_dir = os.path.dirname(json_file)
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)

        # Dump the raw data to json
        with open(json_file, "w+") as f:
            data = {
                'version': self.version,
                'time': self.start_time.isoformat(),
                'tests': self.tests,
                'revision': self.revision,
                'schema': self.schema,
            }
            json.dump(data, f, indent=4)



"""
noseperf.wrappers.redis
~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2012 DISQUS
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import

import time
from noseperf.wrappers.base import Wrapper


class RedisPipelineWrapper(Wrapper):
    _perftype = 'redis'

    def __call__(self, func, pipeline, *args, **kwargs):
        __traceback_hide__ = True  # NOQA

        command_stack = pipeline.command_stack[:]

        start = time.time()
        try:
            return func(pipeline, *args, **kwargs)
        finally:
            end = time.time()

            data = {
                'name': 'pipeline',
                'args': repr(command_stack),
                'kwargs': repr({}),
                'start': start,
                'end': end,
            }

            self._record(data)


class RedisWrapper(Wrapper):
    _perftype = 'redis'

    def __call__(self, func, *args, **kwargs):
        __traceback_hide__ = True  # NOQA

        start = time.time()
        try:
            return func(*args, **kwargs)
        finally:
            end = time.time()

            data = {
                'name': args[1],
                'args': repr(args[2:]),
                'kwargs': repr(kwargs),
                'start': start,
                'end': end,
            }
            self._record(data)



"""
noseperf.wrappers.django
~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2012 DISQUS
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import

import time
from noseperf.wrappers.base import Wrapper


class DjangoTemplateWrapper(Wrapper):
    _perftype = 'template'

    def __call__(self, func, template, *args, **kwargs):
        __traceback_hide__ = True  # NOQA

        start = time.time()

        try:
            return func(template, *args, **kwargs)
        finally:
            end = time.time()
            self._record({
                'name': 'render',
                'args': [template.name],
                'start': start,
                'end': end,
            })


class DjangoCursorWrapper(Wrapper):
    _perftype = 'sql'

    def __init__(self, cursor, connection, *args, **kwargs):
        self.cursor = cursor
        self.connection = connection
        super(DjangoCursorWrapper, self).__init__(*args, **kwargs)

    def __getattr__(self, attr):
        __traceback_hide__ = True  # NOQA

        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            return getattr(self.cursor, attr)

    def __iter__(self):
        __traceback_hide__ = True  # NOQA

        return iter(self.cursor)

    def execute(self, operation, parameters=()):
        """
        Wraps execute method to record the query, execution duration and
        stackframe.
        """
        __traceback_hide__ = True  # NOQ

        # Time the exection of the query

        start = time.time()
        try:
            return self.cursor.execute(operation, parameters)
        finally:
            end = time.time()

            # Save the data
            data = {
                'name': operation,
                'args': parameters,
                'start': start,
                'end': end,
            }
            self._record(data)

    def executemany(self, operation, parameters):
        __traceback_hide__ = True  # NOQA

        # The real cursor calls execute in executemany so nothing needs to be
        # recorded here.
        self.cursor.executemany(operation, parameters)


def patch_cursor(data):
    def patched_cursor(func, self):
        cursor = func(self)
        return DjangoCursorWrapper(cursor, self, data)
    return patched_cursor



"""
noseperf.util
~~~~~~~~~~~~~

:copyright: (c) 2012 DISQUS
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import

from functools import wraps


def _dot_lookup(thing, comp, import_path):
    try:
        return getattr(thing, comp)
    except AttributeError:
        __import__(import_path)
        return getattr(thing, comp)


def import_string(target):
    components = target.split('.')
    import_path = components.pop(0)
    thing = __import__(import_path)

    for comp in components:
        import_path += ".%s" % comp
        thing = _dot_lookup(thing, comp, import_path)
    return thing


class PatchContext(object):
    def __init__(self, target, callback):
        target, attr = target.rsplit('.', 1)
        target = import_string(target)
        self.func = getattr(target, attr)
        self.target = target
        self.attr = attr
        self.callback = callback

    def __enter__(self):
        @wraps(getattr(self.target, self.attr))
        def wrapped(*args, **kwargs):
            __traceback_hide__ = True  # NOQA
            return self.callback(self.func, *args, **kwargs)

        setattr(self.target, self.attr, wrapped)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        setattr(self.target, self.attr, self.func)



"""
noseperf.stacks
~~~~~~~~~~~~~~~

This code is part of the `Raven <https://github.com/getsentry/raven>`_ project.

It's nearly identical except that we don't extract local variables.

:copyright: (c) 2010 by the Sentry Team.
:license: BSD, see LICENSE for more details.
"""

import inspect
import re
import sys


_coding_re = re.compile(r'coding[:=]\s*([-\w.]+)')


def frames_after_module(frames, module):
    started = False
    last_mod = ''
    results = []
    for frame, lineno in list(frames)[::-1]:
        if not started:
            f_globals = getattr(frame, 'f_globals', {})
            module_name = _getitem_from_frame(f_globals, '__name__')
            if (last_mod and last_mod.startswith('unittest2')) \
                and not module_name.startswith('unittest2'):
                started = True
            else:
                last_mod = module_name
                continue
        results.append((frame, lineno))
    return results


def get_lines_from_file(filename, lineno, context_lines, loader=None, module_name=None):
    """
    Returns context_lines before and after lineno from file.
    Returns (pre_context_lineno, pre_context, context_line, post_context).
    """
    source = None
    if loader is not None and hasattr(loader, "get_source"):
        try:
            source = loader.get_source(module_name)
        except ImportError:
            # Traceback (most recent call last):
            #   File "/Users/dcramer/Development/django-sentry/sentry/client/handlers.py", line 31, in emit
            #     get_client().create_from_record(record, request=request)
            #   File "/Users/dcramer/Development/django-sentry/sentry/client/base.py", line 325, in create_from_record
            #     data['__sentry__']['frames'] = varmap(shorten, get_stack_info(stack))
            #   File "/Users/dcramer/Development/django-sentry/sentry/utils/stacks.py", line 112, in get_stack_info
            #     pre_context_lineno, pre_context, context_line, post_context = get_lines_from_file(filename, lineno, 7, loader, module_name)
            #   File "/Users/dcramer/Development/django-sentry/sentry/utils/stacks.py", line 24, in get_lines_from_file
            #     source = loader.get_source(module_name)
            #   File "/System/Library/Frameworks/Python.framework/Versions/2.7/lib/python2.7/pkgutil.py", line 287, in get_source
            #     fullname = self._fix_name(fullname)
            #   File "/System/Library/Frameworks/Python.framework/Versions/2.7/lib/python2.7/pkgutil.py", line 262, in _fix_name
            #     "module %s" % (self.fullname, fullname))
            # ImportError: Loader for module cProfile cannot handle module __main__
            source = None
        if source is not None:
            source = source.splitlines()

    if source is None:
        try:
            f = open(filename)
            try:
                source = f.readlines()
            finally:
                f.close()
        except (OSError, IOError):
            pass

    if source is None:
        return []

    encoding = 'ascii'
    for line in source[:2]:
        # File coding may be specified. Match pattern from PEP-263
        # (http://www.python.org/dev/peps/pep-0263/)
        match = _coding_re.search(line)
        if match:
            encoding = match.group(1)
            break
    source = [unicode(sline, encoding, 'replace') for sline in source]

    lower_bound = max(0, lineno - context_lines)
    upper_bound = min(lineno + context_lines + 1, len(source))

    try:
        return [(lineno + 1, source[lineno].strip('\n')) for lineno in xrange(lower_bound, upper_bound)]
    except IndexError:
        # the file may have changed since it was loaded into memory
        return []


def _getitem_from_frame(f_locals, key, default=None):
    """
    f_locals is not guaranteed to have .get(), but it will always
    support __getitem__. Even if it doesnt, we return ``default``.
    """
    try:
        return f_locals[key]
    except Exception:
        return default


def iter_stack_frames(frames=None):
    """
    Given an optional list of frames (defaults to current stack),
    iterates over all frames that do not contain the ``__traceback_hide__``
    local variable.
    """
    if not frames:
        frames = inspect.stack()[1:]

    for frame, lineno in ((f[0], f[2]) for f in frames):
        f_locals = getattr(frame, 'f_locals', {})
        if _getitem_from_frame(f_locals, '__traceback_hide__'):
            continue
        yield frame, lineno


def get_stack_info(frames):
    """
    Given a list of frames, returns a list of stack information
    dictionary objects that are JSON-ready.

    We have to be careful here as certain implementations of the
    _Frame class do not contain the nescesary data to lookup all
    of the information we want.
    """
    __traceback_hide__ = True  # NOQA

    results = []
    for frame_info in frames:
        # Old, terrible API
        if isinstance(frame_info, (list, tuple)):
            frame, lineno = frame_info

        else:
            frame = frame_info
            lineno = frame_info.f_lineno

        # Support hidden frames
        f_locals = getattr(frame, 'f_locals', {})
        if _getitem_from_frame(f_locals, '__traceback_hide__'):
            continue

        f_globals = getattr(frame, 'f_globals', {})
        loader = _getitem_from_frame(f_globals, '__loader__')
        module_name = _getitem_from_frame(f_globals, '__name__')

        f_code = getattr(frame, 'f_code', None)
        if f_code:
            abs_path = frame.f_code.co_filename
            function = frame.f_code.co_name
        else:
            abs_path = None
            function = None

        if lineno:
            lineno -= 1

        if lineno is not None and abs_path:
            context = get_lines_from_file(abs_path, lineno, 3, loader, module_name)
        else:
            context = []

        # Try to pull a relative file path
        # This changes /foo/site-packages/baz/bar.py into baz/bar.py
        try:
            base_filename = sys.modules[module_name.split('.', 1)[0]].__file__
            filename = abs_path.split(base_filename.rsplit('/', 2)[0], 1)[-1][1:]
        except:
            filename = abs_path

        if not filename:
            filename = abs_path

        frame_result = {
            'abs_path': abs_path,
            'filename': filename,
            'module': module_name,
            'function': function,
            'lineno': lineno + 1,
            'context': context,
        }
        results.append(frame_result)
    return results



"""
noseperf.wrappers.base
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2012 DISQUS
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import

import inspect
import time
from noseperf.stacks import get_stack_info, iter_stack_frames, frames_after_module


class Wrapper(object):
    """Wraps Redis methods and logs the results """

    def __init__(self, data):
        self.data = data

    def _record(self, data):
        frames = frames_after_module(iter_stack_frames(inspect.stack()[2:]), 'unittest2')
        data['stacktrace'] = get_stack_info(frames)
        if 'type' not in data:
            data['type'] = self._perftype
        self.data.append(data)


class FunctionWrapper(Wrapper):
    _perfkey = None

    def __init__(self, perfkey=None, *args, **kwargs):
        super(FunctionWrapper, self).__init__(*args, **kwargs)
        self._perfkey = perfkey

    def __call__(self, func, *args, **kwargs):
        __traceback_hide__ = True  # NOQA

        start = time.time()
        try:
            return func(*args, **kwargs)
        finally:
            end = time.time()

            if getattr(func, 'im_class', None):
                arg_str = repr(args[1:])
            else:
                arg_str = repr(args)

            data = {
                'type': self._perfkey or func.__name__,
                'name': func.__name__,
                'args': arg_str,
                'kwargs': repr(kwargs),
                'start': start,
                'end': end,
            }

            self._record(data)



"""
nose-performance
================

:copyright: (c) 2012 DISQUS
:license: Apache License 2.0, see LICENSE for more details.
"""

from setuptools import setup, find_packages

# Hack to prevent stupid "TypeError: 'NoneType' object is not callable" error
# in multiprocessing/util.py _exit_function when running `python
# setup.py test` (see
# http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html)
try:
    import multiprocessing
except ImportError:
    pass

tests_require = [
    'psycopg2',
    'redis',
    'Django>=1.2',
]

install_requires = [
    'nose>=0.9',
    'simplejson',
    'unittest2',
]

setup(
    name='nose-performance',
    version='0.4.1',
    author='DISQUS',
    author_email='opensource@disqus.com',
    url='https://github.com/disqus/nose-performance',
    description='A plugin for Nose for running performance tests',
    long_description=__doc__,
    package_dir={'': 'src'},
    packages=find_packages('src'),
    zip_safe=False,
    install_requires=install_requires,
    # tests_require=tests_require,
    # test_suite='runtests.runtests',
    license='Apache License 2.0',
    include_package_data=True,
    entry_points={
       'nose.plugins.0.10': [
            'noseperf = noseperf.plugin:PerformancePlugin'
        ]
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Software Development'
    ],
)
